package lcsb.mapviewer.api.projects.models;

import java.awt.geom.Point2D;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;

import org.primefaces.model.map.LatLng;

import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.model.map.layout.Layout;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.map.model.ModelSubmodelConnection;
import lcsb.mapviewer.model.map.model.SubmodelType;
import lcsb.mapviewer.services.utils.gmap.CoordinationConverter;
import lcsb.mapviewer.services.view.LayoutView;
import lcsb.mapviewer.services.view.LayoutViewFactory;

public class ModelMetaData implements Serializable {

  /**
   * 
   */
  private static final long serialVersionUID = 1L;

  /**
   * Version of the model.
   */
  private String version;

  /**
   * Name of the model.
   */
  private String name;

  private Integer idObject;

  /**
   * Size in pixels of the single square tile (small image used for graphical
   * representation).
   */
  private Integer tileSize;

  private Integer width;

  private Integer height;

  private Double defaultCenterX;

  private Double defaultCenterY;
  
  private Integer defaultZoomLevel;

  /**
   * Minimum zoom level that should be allowed by the Google Maps API.
   */
  private Integer minZoom;

  /**
   * Maximum zoom level that should be allowed by the Google Maps API.
   */
  private Integer maxZoom;

  private List<LayoutView> layouts = new ArrayList<>();

  /**
   * List of submodels in the model.
   */
  private List<ModelMetaData> submodels = new ArrayList<>();

  /**
   * Where is the center of the map in latitude, longitude format.
   */
  private LatLng centerLatLng;

  /**
   * Top-Left corner of the map (0,0) as a latLng coordinates.
   */
  private LatLng topLeftLatLng;

  /**
   * Bottom-Right corner of the map (width,height) as a latLng coordinates.
   */
  private LatLng bottomRightLatLng;

  /**
   * Type of the submodel.
   */
  private SubmodelType submodelType = SubmodelType.UNKNOWN;

  /**
   * Default constructor.
   */
  public ModelMetaData(Model model) {
    this(model.getModelData());
  }

  public ModelMetaData(ModelData model) {
    this.setName(model.getName());
    this.setIdObject(model.getId());

    this.setMinZoom(Configuration.MIN_ZOOM_LEVEL);
    this.setMaxZoom(this.getMinZoom() + model.getZoomLevels());
    this.setTileSize(model.getTileSize());
    this.setWidth((int) (double) model.getWidth());
    this.setHeight((int) (double) model.getHeight());
    int size = Math.max(width, height);
    CoordinationConverter cConverter = new CoordinationConverter(model);
    this.setCenterLatLng(cConverter.toLatLng(new Point2D.Double(size / 2, size / 2)));
    this.setBottomRightLatLng(cConverter.toLatLng(new Point2D.Double(model.getWidth(), model.getHeight())));
    this.setTopLeftLatLng(cConverter.toLatLng(new Point2D.Double(0, 0)));
    this.setDefaultCenterX(model.getDefaultCenterX());
    this.setDefaultCenterY(model.getDefaultCenterY());
    this.setDefaultZoomLevel(model.getDefaultZoomLevel());

    List<ModelMetaData> submodels = new ArrayList<>();
    for (ModelSubmodelConnection connection : model.getSubmodels()) {
      ModelMetaData submodelData = new ModelMetaData(connection.getSubmodel());
      submodelData.setSubmodelType(connection.getType());
      submodels.add(submodelData);
    }
    LayoutViewFactory factory = new LayoutViewFactory();
    for (Layout layout : model.getLayouts()) {
      if (layout.isPublicLayout()) {
        layouts.add(factory.create(layout));
      }
    }
    this.setSubmodels(submodels);

  }

  protected ModelMetaData() {
  }

  /**
   * @return the version
   * @see #version
   */
  public String getVersion() {
    return version;
  }

  /**
   * @param version
   *          the version to set
   * @see #version
   */
  public void setVersion(String version) {
    this.version = version;
  }

  /**
   * @return the tileSize
   * @see #tileSize
   */
  public Integer getTileSize() {
    return tileSize;
  }

  /**
   * @param tileSize
   *          the tileSize to set
   * @see #tileSize
   */
  public void setTileSize(Integer tileSize) {
    this.tileSize = tileSize;
  }

  /**
   * @return the minZoom
   * @see #minZoom
   */
  public Integer getMinZoom() {
    return minZoom;
  }

  /**
   * @param minZoom
   *          the minZoom to set
   * @see #minZoom
   */
  public void setMinZoom(Integer minZoom) {
    this.minZoom = minZoom;
  }

  /**
   * @return the maxZoom
   * @see #maxZoom
   */
  public Integer getMaxZoom() {
    return maxZoom;
  }

  /**
   * @param maxZoom
   *          the maxZoom to set
   * @see #maxZoom
   */
  public void setMaxZoom(Integer maxZoom) {
    this.maxZoom = maxZoom;
  }

  /**
   * @return the centerLatLng
   * @see #centerLatLng
   */
  public LatLng getCenterLatLng() {
    return centerLatLng;
  }

  /**
   * @param centerLatLng
   *          the centerLatLng to set
   * @see #centerLatLng
   */
  public void setCenterLatLng(LatLng centerLatLng) {
    this.centerLatLng = centerLatLng;
  }

  /**
   * @return the name
   * @see #name
   */
  public String getName() {
    return name;
  }

  /**
   * @param name
   *          the name to set
   * @see #name
   */
  public void setName(String name) {
    this.name = name;
  }

  /**
   * @return the topLeftLatLng
   * @see #topLeftLatLng
   */
  public LatLng getTopLeftLatLng() {
    return topLeftLatLng;
  }

  /**
   * @param topLeftLatLng
   *          the topLeftLatLng to set
   * @see #topLeftLatLng
   */
  public void setTopLeftLatLng(LatLng topLeftLatLng) {
    this.topLeftLatLng = topLeftLatLng;
  }

  /**
   * @return the bottomRightLatLng
   * @see #bottomRightLatLng
   */
  public LatLng getBottomRightLatLng() {
    return bottomRightLatLng;
  }

  /**
   * @param bottomRightLatLng
   *          the bottomRightLatLng to set
   * @see #bottomRightLatLng
   */
  public void setBottomRightLatLng(LatLng bottomRightLatLng) {
    this.bottomRightLatLng = bottomRightLatLng;
  }

  /**
   * @return the width
   * @see #width
   */
  public Integer getWidth() {
    return width;
  }

  /**
   * @param width
   *          the width to set
   * @see #width
   */
  public void setWidth(Integer width) {
    this.width = width;
  }

  /**
   * @return the height
   * @see #height
   */
  public Integer getHeight() {
    return height;
  }

  /**
   * @param height
   *          the height to set
   * @see #height
   */
  public void setHeight(Integer height) {
    this.height = height;
  }

  /**
   * @return the submodels
   * @see #submodels
   */
  public List<ModelMetaData> getSubmodels() {
    return submodels;
  }

  /**
   * @param submodels
   *          the submodels to set
   * @see #submodels
   */
  public void setSubmodels(List<ModelMetaData> submodels) {
    this.submodels = submodels;
  }

  /**
   * @return the layouts
   * @see #layouts
   */
  public List<LayoutView> getLayouts() {
    return layouts;
  }

  /**
   * @param layouts
   *          the layouts to set
   * @see #layouts
   */
  public void setLayouts(List<LayoutView> layouts) {
    this.layouts = layouts;
  }

  /**
   * @return the idObject
   * @see #idObject
   */
  public Integer getIdObject() {
    return idObject;
  }

  /**
   * @param idObject
   *          the idObject to set
   * @see #idObject
   */
  public void setIdObject(Integer idObject) {
    this.idObject = idObject;
  }

  /**
   * @return the submodelType
   * @see #submodelType
   */
  public SubmodelType getSubmodelType() {
    return submodelType;
  }

  /**
   * @param submodelType
   *          the submodelType to set
   * @see #submodelType
   */
  public void setSubmodelType(SubmodelType submodelType) {
    this.submodelType = submodelType;
  }

  public Integer getDefaultZoomLevel() {
    return defaultZoomLevel;
  }

  public void setDefaultZoomLevel(Integer defaultZoomLevel) {
    this.defaultZoomLevel = defaultZoomLevel;
  }

  public Double getDefaultCenterX() {
    return defaultCenterX;
  }

  public void setDefaultCenterX(Double defaultCenterX) {
    this.defaultCenterX = defaultCenterX;
  }

  public Double getDefaultCenterY() {
    return defaultCenterY;
  }

  public void setDefaultCenterY(Double defaultCenterY) {
    this.defaultCenterY = defaultCenterY;
  }

}
