package lcsb.mapviewer.api.projects.models;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.api.ObjectNotFoundException;
import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.common.exception.InvalidArgumentException;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.model.ModelData;
import lcsb.mapviewer.model.user.PrivilegeType;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.interfaces.ILayoutService;
import lcsb.mapviewer.services.view.AuthenticationToken;

@Transactional(value = "txManager")
public class ModelRestImpl extends BaseRestImpl {

  @Autowired
  private ILayoutService layoutService;

  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private Logger logger = Logger.getLogger(ModelRestImpl.class);

  public List<ModelMetaData> getModels(String projectId, String token)
      throws SecurityException, ObjectNotFoundException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Project project = getProjectService().getProjectByProjectId(projectId, authenticationToken);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    List<ModelMetaData> result = createData(project, authenticationToken);
    return result;
  }

  public ModelMetaData getModel(String projectId, String modelId, String token) throws SecurityException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    Model submodel = model.getSubmodelById(modelId);
    if (submodel == null) {
      return null;
    } else {
      ModelMetaData result = new ModelMetaData(submodel);

      List<ModelMetaData> metaDataToProcess = new ArrayList<>();
      metaDataToProcess.add(result);
      metaDataToProcess.addAll(result.getSubmodels());
      for (ModelMetaData submodelData : metaDataToProcess) {
        submodelData.setLayouts(layoutService.getGeneralLayouts(model.getSubmodelById(submodelData.getIdObject())));
        logger.debug(submodelData.getLayouts());
      }
      return result;
    }
  }

  private List<ModelMetaData> createData(Project project, AuthenticationToken token) throws SecurityException {
    List<ModelMetaData> result = new ArrayList<>();
    Model model = getModelService().getLastModelByProjectId(project.getProjectId(), token);
    if (model != null) {

      List<Model> models = new ArrayList<>();
      models.add(model);
      models.addAll(model.getSubmodels());
      for (Model model2 : models) {
        result.add(getModel(project.getProjectId(), model2.getId() + "", token.getId()));
      }
    }
    return result;
  }

  public ModelMetaData updateModel(String projectId, String modelId, Map<String, Object> data, String token)
      throws SecurityException, QueryException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Project project = getProjectService().getProjectByProjectId(projectId, authenticationToken);
    if (project == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    boolean canModify = getUserService().userHasPrivilege(authenticationToken, PrivilegeType.ADD_MAP);
    if (!canModify) {
      throw new SecurityException("You cannot update projects");
    }
    ModelData model = null;
    Integer id = Integer.valueOf(modelId);
    for (ModelData m : project.getModels()) {
      if (m.getId().equals(id)) {
        model = m;
      }
    }
    if (model == null) {
      throw new ObjectNotFoundException("Model with given id doesn't exist");
    }

    Set<String> fields = data.keySet();
    for (String fieldName : fields) {
      Object value = data.get(fieldName);
      if (fieldName.equalsIgnoreCase("defaultCenterX")) {
        model.setDefaultCenterX(parseDouble(value));
      } else if (fieldName.equalsIgnoreCase("defaultCenterY")) {
        model.setDefaultCenterY(parseDouble(value));
      } else if (fieldName.equalsIgnoreCase("defaultZoomLevel")) {
        model.setDefaultZoomLevel(parseInteger(value));
      } else if (fieldName.equalsIgnoreCase("id")) {
        if (!model.getId().equals(parseInteger(value))) {
          throw new QueryException("Id doesn't match: " + value + ", " + model.getId());
        }
      } else {
        throw new QueryException("Unknown field: " + fieldName);
      }
    }
    getModelService().updateModel(model, authenticationToken);
    return getModel(projectId, modelId, token);
  }

  private Integer parseInteger(Object value) throws QueryException {
    if (value instanceof Integer) {
      return (Integer) value;
    } else if (value instanceof String) {
      if (((String) value).equalsIgnoreCase("")) {
        return null;
      } else {
        try {
          return Integer.parseInt((String) value);
        } catch (NumberFormatException e) {
          throw new QueryException("Invalid double value: " + value);
        }
      }
    } else {
      throw new InvalidArgumentException();
    }

  }

  private Double parseDouble(Object value) throws QueryException {
    if (value instanceof Double) {
      return (Double) value;
    } else if (value instanceof String) {
      if (((String) value).equalsIgnoreCase("")) {
        return null;
      } else {
        try {
          return Double.parseDouble((String) value);
        } catch (NumberFormatException e) {
          throw new QueryException("Invalid double value: " + value);
        }
      }
    } else if (value == null) {
      return null;
    } else {
      throw new InvalidArgumentException();
    }
  }
}
