package lcsb.mapviewer.api.projects.models.bioEntities.reactions;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.xml.transform.TransformerException;

import org.apache.commons.lang3.StringUtils;
import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.common.exception.InvalidXmlSchemaException;
import lcsb.mapviewer.model.map.kinetics.SbmlFunction;
import lcsb.mapviewer.model.map.kinetics.SbmlKinetics;
import lcsb.mapviewer.model.map.kinetics.SbmlParameter;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.reaction.Modifier;
import lcsb.mapviewer.model.map.reaction.Product;
import lcsb.mapviewer.model.map.reaction.Reactant;
import lcsb.mapviewer.model.map.reaction.Reaction;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.UserAccessException;
import lcsb.mapviewer.services.interfaces.IModelService;
import lcsb.mapviewer.services.interfaces.IUserService;
import lcsb.mapviewer.services.search.data.LightReactionView;

@Transactional(value = "txManager")
public class ReactionsRestImpl extends BaseRestImpl {

  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private Logger logger = Logger.getLogger(ReactionsRestImpl.class);

  @Autowired
  private IUserService userService;

  @Autowired
  private IModelService modelService;

  /**
   * @return the userService
   * @see #userService
   */
  public IUserService getUserService() {
    return userService;
  }

  /**
   * @param userService
   *          the userService to set
   * @see #userService
   */
  public void setUserService(IUserService userService) {
    this.userService = userService;
  }

  public List<Map<String, Object>> getReactions(String projectId, String id, String columns, String modelId,
      String token, String participantElementId) throws UserAccessException, SecurityException {
    Set<Integer> ids = new HashSet<>();
    if (!id.equals("")) {
      for (String str : id.split(",")) {
        ids.add(Integer.valueOf(str));
      }
    }
    List<Model> models = getModels(projectId, modelId, token);

    Set<Element> elementSet = new HashSet<>();
    if (!participantElementId.equals("")) {
      for (String str : participantElementId.split(",")) {
        for (Model model : models) {
          elementSet.add(model.getElementByDbId(Integer.valueOf(str)));
        }
      }
    }
    Set<String> columnsSet = createReactionColumnSet(columns);

    List<Map<String, Object>> result = new ArrayList<>();

    for (Model model2 : models) {
      for (Reaction reaction : model2.getReactions()) {
        if (ids.size() == 0 || ids.contains(reaction.getId())) {
          if (elementSet.size() == 0 || reactionContainsElement(reaction, elementSet)) {
            result.add(preparedReaction(reaction, columnsSet));
          }
        }
      }
    }

    return result;
  }

  private boolean reactionContainsElement(Reaction reaction, Set<Element> elementSet) {
    for (Element element : elementSet) {
      if (reaction.containsElement(element)) {
        return true;
      }
    }
    return false;
  }

  private Map<String, Object> preparedReaction(Reaction reaction, Set<String> columnsSet) {
    Map<String, Object> result = new HashMap<>();
    for (String string : columnsSet) {
      String column = string.toLowerCase();
      Object value = null;
      if (column.equals("id") || column.equals("idobject")) {
        value = reaction.getId();
      } else if (column.equals("modelid")) {
        value = reaction.getModelData().getId();
      } else if (column.equals("reactionid")) {
        value = reaction.getIdReaction();
      } else if (column.equals("name")) {
        value = reaction.getName();
      } else if (column.equals("centerpoint")) {
        value = reaction.getCenterPoint();
      } else if (column.equals("products")) {
        List<Integer> ids = new ArrayList<>();
        for (Product product : reaction.getProducts()) {
          ids.add(product.getElement().getId());
        }
        value = StringUtils.join(ids, ",");
      } else if (column.equals("reactants")) {
        List<Integer> ids = new ArrayList<>();
        for (Reactant reactant : reaction.getReactants()) {
          ids.add(reactant.getElement().getId());
        }
        value = StringUtils.join(ids, ",");
      } else if (column.equals("modifiers")) {
        List<Integer> ids = new ArrayList<>();
        for (Modifier product : reaction.getModifiers()) {
          ids.add(product.getElement().getId());
        }
        value = StringUtils.join(ids, ",");
      } else if (column.equals("type")) {
        value = reaction.getStringType();
      } else if (column.equals("hierarchyvisibilitylevel")) {
        value = reaction.getVisibilityLevel();
      } else if (column.equals("lines")) {
        value = new LightReactionView(reaction).getLines();
      } else if (column.equals("notes")) {
        value = reaction.getNotes();
      } else if (column.equals("kineticlaw")) {
        value = kineticsToMap(reaction.getKinetics());
      } else if (column.equals("references")) {
        value = createAnnotations(reaction.getMiriamData());
      } else {
        value = "Unknown column";
      }
      result.put(string, value);
    }
    return result;
  }

  private Map<String, Object> kineticsToMap(SbmlKinetics kinetics) {
    if (kinetics == null) {
      return null;
    }
    Map<String, Object> result = new HashMap<>();
    result.put("definition", kinetics.getDefinition());
    try {
      result.put("mathMlPresentation", super.mathMLToPresentationML(kinetics.getDefinition()));
    } catch (IOException | InvalidXmlSchemaException | TransformerException e) {
      logger.error("Problems with transforming kinetics", e);
    }

    List<Integer> functionIds = new ArrayList<>();
    for (SbmlFunction function : kinetics.getFunctions()) {
      functionIds.add(function.getId());
    }
    result.put("functionIds", functionIds);
    List<Integer> parameterIds = new ArrayList<>();
    for (SbmlParameter parameter : kinetics.getParameters()) {
      parameterIds.add(parameter.getId());
    }
    result.put("parameterIds", parameterIds);
    return result;
  }

  private Set<String> createReactionColumnSet(String columns) {
    Set<String> columnsSet = new HashSet<>();
    if (columns.equals("")) {
      columnsSet.add("id");
      columnsSet.add("reactionId");
      columnsSet.add("modelId");
      columnsSet.add("type");
      columnsSet.add("lines");
      columnsSet.add("kineticLaw");
      columnsSet.add("centerPoint");
      columnsSet.add("products");
      columnsSet.add("reactants");
      columnsSet.add("modifiers");
      columnsSet.add("hierarchyVisibilityLevel");
      columnsSet.add("references");
      columnsSet.add("notes");
    } else {
      for (String str : columns.split(",")) {
        columnsSet.add(str);
      }
    }
    return columnsSet;
  }

  /**
   * @return the modelService
   * @see #modelService
   */
  public IModelService getModelService() {
    return modelService;
  }

  /**
   * @param modelService
   *          the modelService to set
   * @see #modelService
   */
  public void setModelService(IModelService modelService) {
    this.modelService = modelService;
  }

}
