package lcsb.mapviewer.api.projects.overlays;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.api.BaseRestImpl;
import lcsb.mapviewer.api.ObjectNotFoundException;
import lcsb.mapviewer.api.QueryException;
import lcsb.mapviewer.common.Configuration;
import lcsb.mapviewer.model.cache.FileEntry;
import lcsb.mapviewer.model.cache.UploadedFileEntry;
import lcsb.mapviewer.model.map.layout.InvalidColorSchemaException;
import lcsb.mapviewer.model.map.layout.Layout;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.user.PrivilegeType;
import lcsb.mapviewer.model.user.User;
import lcsb.mapviewer.persist.dao.cache.UploadedFileEntryDao;
import lcsb.mapviewer.persist.dao.map.LayoutDao;
import lcsb.mapviewer.services.SecurityException;
import lcsb.mapviewer.services.interfaces.ILayoutService;
import lcsb.mapviewer.services.interfaces.ILayoutService.CreateLayoutParams;
import lcsb.mapviewer.services.search.data.ElementIdentifier.ElementIdentifierType;
import lcsb.mapviewer.services.search.layout.FullLayoutAliasView;
import lcsb.mapviewer.services.search.layout.FullLayoutReactionView;
import lcsb.mapviewer.services.search.layout.LightLayoutAliasView;
import lcsb.mapviewer.services.search.layout.LightLayoutReactionView;
import lcsb.mapviewer.services.utils.data.ColorSchemaType;
import lcsb.mapviewer.services.view.AuthenticationToken;
import lcsb.mapviewer.services.view.LayoutView;

@Transactional(value = "txManager")
public class OverlayRestImpl extends BaseRestImpl {

  /**
   * Default class logger.
   */
  @SuppressWarnings("unused")
  private Logger logger = Logger.getLogger(OverlayRestImpl.class);

  @Autowired
  private ILayoutService layoutService;

  @Autowired
  private UploadedFileEntryDao uploadedFileEntryDao;

  @Autowired
  private LayoutDao layoutDao;

  public List<LayoutView> getOverlayList(String token, String projectId, String creatorLogin, String publicOverlay)
      throws SecurityException, QueryException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (model == null) {
      return new ArrayList<>();
    }
    User creator = null;
    if (creatorLogin != null && !creatorLogin.isEmpty()) {
      creator = getUserService().getUserByLogin(creatorLogin);
      if (creator == null) {
        throw new ObjectNotFoundException("User with given id doesn't exist: " + creatorLogin);
      }
    }
    Boolean publicData = null;
    if (publicOverlay != null && !publicOverlay.isEmpty()) {
      publicData = publicOverlay.equalsIgnoreCase("true");
    }
    return layoutService.getCustomLayouts(model, token, publicData, creator);
  }

  /**
   * @return the layoutService
   * @see #layoutService
   */
  public ILayoutService getLayoutService() {
    return layoutService;
  }

  /**
   * @param layoutService
   *          the layoutService to set
   * @see #layoutService
   */
  public void setLayoutService(ILayoutService layoutService) {
    this.layoutService = layoutService;
  }

  public List<Map<String, Object>> getOverlayElements(String token, String projectId, int overlayId, String columns)
      throws QueryException, SecurityException {
    List<Map<String, Object>> result = new ArrayList<>();
    AuthenticationToken authenticationToken = getUserService().getToken(token);

    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (model == null) {
      throw new QueryException("Project with given id doesn't exist");
    }
    List<LightLayoutAliasView> speciesList = layoutService.getAliasesForLayout(model, overlayId, authenticationToken);
    for (LightLayoutAliasView lightLayoutAliasView : speciesList) {
      Map<String, Object> element = new HashMap<>();
      element.put("type", ElementIdentifierType.ALIAS);
      element.put("overlayContent", lightLayoutAliasView);
      result.add(element);
    }

    List<LightLayoutReactionView> reactions = layoutService.getReactionsForLayout(model, overlayId,
        authenticationToken);
    for (LightLayoutReactionView lightReactionView : reactions) {
      Map<String, Object> element = new HashMap<>();
      element.put("type", ElementIdentifierType.REACTION);
      element.put("overlayContent", lightReactionView);
      result.add(element);
    }
    return result;
  }

  public LayoutView getOverlayById(String token, String projectId, String overlayId)
      throws SecurityException, QueryException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (model == null) {
      throw new QueryException("Project with given id doesn't exist");
    }
    return layoutService.getLayoutById(Integer.valueOf(overlayId), authenticationToken);
  }

  public FileEntry getOverlaySource(String token, String projectId, String overlayId)
      throws SecurityException, QueryException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (model == null) {
      throw new QueryException("Project with given id doesn't exist");
    }
    try {
      Integer id = Integer.valueOf(overlayId);
      Layout layout = layoutService.getLayoutDataById(id, authenticationToken);
      if (layout == null) {
        throw new QueryException("Invalid overlay id");
      }
      // lazy initialization issue
      layout.getInputData().getFileContent();
      return layout.getInputData();
    } catch (NumberFormatException e) {
      throw new QueryException("Invalid overlay id");
    }
  }

  public LayoutView updateOverlay(String token, String projectId, String overlayId, Map<String, Object> overlayData)
      throws QueryException, SecurityException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    if (overlayData == null) {
      throw new QueryException("overlay field cannot be undefined");
    }
    try {
      Integer id = Integer.valueOf(overlayId);
      Layout layout = layoutService.getLayoutDataById(id, authenticationToken);
      if (layout == null) {
        throw new ObjectNotFoundException("overlay doesn't exist");
      }
      boolean isAdmin = getUserService().userHasPrivilege(authenticationToken, PrivilegeType.LAYOUT_MANAGEMENT,
          layout.getModel().getProject());
      if (layout.isPublicLayout() && !isAdmin) {
        throw new SecurityException("You cannot modify given overlay");
      }
      for (String key : overlayData.keySet()) {
        Object value = overlayData.get(key);
        if (key.equalsIgnoreCase("description")) {
          layout.setDescription((String) value);
        } else if (key.equalsIgnoreCase("name")) {
          layout.setTitle((String) value);
        } else if (key.equalsIgnoreCase("publicOverlay")) {
          layout.setPublicLayout(parseBoolean(value));
        } else if (key.equalsIgnoreCase("defaultOverlay")) {
          layout.setDefaultOverlay(parseBoolean(value));
        } else if (key.equalsIgnoreCase("creator")) {
          if ("".equals(value)) {
            layout.setCreator(null);
          } else {
            layout.setCreator(getUserService().getUserByLogin((String) value));
          }
        } else {
          throw new QueryException("Unknown parameter: " + key);
        }
      }
      layoutDao.update(layout);
      return layoutService.getLayoutById(Integer.valueOf(overlayId), authenticationToken);
    } catch (NumberFormatException e) {
      throw new ObjectNotFoundException("overlay doesn't exist");
    }
  }

  private boolean parseBoolean(Object value) {
    if (value instanceof Boolean) {
      return (Boolean) value;
    } else {
      return "true".equalsIgnoreCase((String) value);
    }
  }

  public Map<String, Object> removeOverlay(String token, String projectId, String overlayId)
      throws QueryException, SecurityException, IOException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (model == null) {
      throw new ObjectNotFoundException("Project with given id doesn't exist");
    }
    try {
      Integer id = Integer.valueOf(overlayId);
      LayoutView layout = layoutService.getLayoutById(id, authenticationToken);
      if (layout == null) {
        throw new ObjectNotFoundException("Overlay doesn't exist");
      }
      if (layoutService.userCanRemoveLayout(layout, authenticationToken)) {
        layoutService.removeLayout(layout, null);
        return okStatus();
      } else {
        throw new SecurityException("You cannot remove given overlay");
      }
    } catch (NumberFormatException e) {
      throw new QueryException("Invalid overlay id: " + overlayId);
    }
  }

  /**
   * @return the layoutDao
   * @see #layoutDao
   */
  public LayoutDao getLayoutDao() {
    return layoutDao;
  }

  /**
   * @param layoutDao
   *          the layoutDao to set
   * @see #layoutDao
   */
  public void setLayoutDao(LayoutDao layoutDao) {
    this.layoutDao = layoutDao;
  }

  public LayoutView addOverlay(String token, String projectId, String name, String description, String content,
      String fileId, String filename, String type) throws SecurityException, QueryException, IOException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);
    User user = getUserService().getUserByToken(token);
    if (Configuration.ANONYMOUS_LOGIN.equals(user.getLogin())) {
      throw new SecurityException("You have no privileges to add overlay");
    }
    Model model = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (model == null) {
      throw new QueryException("Project with given id doesn't exist");
    }
    ColorSchemaType colorSchemaType = ColorSchemaType.GENERIC;
    if (type != null && !type.equals("")) {
      try {
        colorSchemaType = ColorSchemaType.valueOf(type);
      } catch (IllegalArgumentException e) {
        throw new QueryException("Invalid type of overlay: " + type, e);
      }
    }
    if (content.isEmpty() && fileId.isEmpty()) {
      throw new QueryException("Either content or fileId must be provided");
    }

    try {
      InputStream stream = null;
      if (!content.isEmpty()) {
        stream = new ByteArrayInputStream(content.getBytes(StandardCharsets.UTF_8));
      } else {
        try {
          int id = Integer.valueOf(fileId);
          UploadedFileEntry file = uploadedFileEntryDao.getById(id);
          if (file == null) {
            throw new QueryException("Invalid file id: " + fileId);
          }
          if (file.getOwner() == null || !file.getOwner().getLogin().equals(user.getLogin())) {
            throw new SecurityException("Access denied to source file");
          }
          stream = new ByteArrayInputStream(file.getFileContent());
        } catch (NumberFormatException e) {
          throw new QueryException("Invalid fileId: " + fileId);
        }
      }

      LayoutView layout = layoutService.createLayout(new CreateLayoutParams().async(false).colorInputStream(stream)
          .description(description).layoutFileName(filename).model(model).name(name).user(user)
          .colorSchemaType(colorSchemaType).directory("."));
      return layoutService.getLayoutById(Integer.valueOf(layout.getIdObject()), authenticationToken);
    } catch (InvalidColorSchemaException e) {
      throw new QueryException(e.getMessage(), e);
    }

  }

  public Map<String, Object> getOverlayElement(String token, String projectId, Integer modelId, Integer overlayId,
      Integer elementId, String elementType, String columns) throws QueryException, SecurityException {
    AuthenticationToken authenticationToken = getUserService().getToken(token);

    Model topModel = getModelService().getLastModelByProjectId(projectId, authenticationToken);
    if (topModel == null) {
      throw new QueryException("Project with given id doesn't exist");
    }
    Model model = topModel.getSubmodelById(modelId);

    Map<String, Object> result = new HashMap<>();
    if (ElementIdentifierType.ALIAS.getJsName().equals(elementType)) {
      FullLayoutAliasView layoutAliasView = layoutService.getFullAliasForLayout(model, elementId, overlayId,
          authenticationToken);
      result.put("type", ElementIdentifierType.ALIAS);
      result.put("overlayContent", layoutAliasView);
      return result;
    } else if (ElementIdentifierType.REACTION.getJsName().equals(elementType)) {
      FullLayoutReactionView layoutAliasView = layoutService.getFullReactionForLayout(model, elementId, overlayId,
          authenticationToken);
      result.put("type", ElementIdentifierType.REACTION);
      result.put("overlayContent", layoutAliasView);
      return result;
    } else {
      throw new QueryException("Unknown element type: " + elementType);
    }
  }

}
